@@ -26,6 +26,7 @@ site: http://bugseng.com/products/ppl/ . */
 #include "Grid.defs.hh"
 #include "Grid_Generator.defs.hh"
 #include "Scalar_Products.defs.hh"
+#include "Scalar_Products.inlines.hh"
 #include "assert.hh"
 #include <string>
 #include <iostream>
@@ -55,49 +56,47 @@ PPL::Grid::construct(dimension_type num_dimensions,
   if (kind == EMPTY) {
     // Set emptiness directly instead of with set_empty, as gen_sys is
     // already correctly initialized.
-
     status.set_empty();
 
     // Extend the zero dim false congruence system to the appropriate
     // dimension and then store it in `con_sys'.
     Congruence_System cgs(Congruence::zero_dim_false());
-    cgs.increase_space_dimension(space_dim);
-    con_sys.m_swap(cgs);
+    cgs.set_space_dimension(space_dim);
+    swap(con_sys, cgs);
 
     PPL_ASSERT(OK());
     return;
   }
 
-  if (num_dimensions > 0) {
-    con_sys.increase_space_dimension(num_dimensions);
-
-    // Initialize both systems to universe representations.
-
-    set_congruences_minimized();
-    set_generators_minimized();
-    dim_kinds.resize(num_dimensions + 1);
-
-    // Extend the zero dim integrality congruence system to the
-    // appropriate dimension and then store it in `con_sys'.
-    Congruence_System cgs(Congruence::zero_dim_integrality());
-    cgs.increase_space_dimension(space_dim);
-    cgs[0][0] = 1; // Recover minimal form after cgs(zdi) normalization.
-    con_sys.m_swap(cgs);
-
-    dim_kinds[0] = PROPER_CONGRUENCE /* a.k.a. PARAMETER */;
-
-    // Trivially true point.
-    gen_sys.insert(grid_point(0*(Variable(0))));
-
-    // A line for each dimension.
-    dimension_type dim = 0;
-    while (dim < num_dimensions) {
-      gen_sys.insert(grid_line(Variable(dim)));
-      dim_kinds[++dim] = CON_VIRTUAL /* a.k.a. LINE */;
-    }
-  }
-  else
+  if (space_dim == 0) {
     set_zero_dim_univ();
+    return;
+  }
+
+  // Initialize both systems to universe representations.
+  set_congruences_minimized();
+  set_generators_minimized();
+  dim_kinds.resize(space_dim + 1);
+
+  // Building a universe congruence system.
+  // Extend the zero dim integrality congruence system to the
+  // appropriate dimension and then store it in `con_sys'.
+  Congruence_System cgs(Congruence::zero_dim_integrality());
+  cgs.set_space_dimension(num_dimensions);
+  // Recover minimal form after cgs(zdi) normalization.
+  cgs.rows[0].expr.set_inhomogeneous_term(Coefficient_one());
+  PPL_ASSERT(cgs.OK());
+  swap(con_sys, cgs);
+
+  // Building a universe grid generator system (and dim_kinds).
+  gen_sys.set_space_dimension(space_dim);
+  gen_sys.insert(grid_point());
+  dim_kinds[0] = PROPER_CONGRUENCE /* a.k.a. PARAMETER */;
+  for (dimension_type dim = 0; dim < space_dim; ++dim) {
+    gen_sys.insert(grid_line(Variable(dim)));
+    dim_kinds[1+dim] = CON_VIRTUAL /* a.k.a. LINE */;
+  }
+  PPL_ASSERT(OK());
 }
 
 void
@@ -121,18 +120,17 @@ PPL::Grid::construct(Congruence_System& cgs) {
   }
   else {
     // Here `space_dim == 0'.
-    if (cgs.num_columns() > 1)
-      // See if an inconsistent congruence has been passed.
-      for (dimension_type i = cgs.num_rows(); i-- > 0; )
-	if (cgs[i].is_inconsistent()) {
-	  // Inconsistent congruence found: the grid is empty.
-	  status.set_empty();
-	  // Insert the zero dim false congruence system into `con_sys'.
-	  // `gen_sys' is already in empty form.
-	  con_sys.insert(Congruence::zero_dim_false());
-	  PPL_ASSERT(OK());
-	  return;
-	}
+    // See if an inconsistent congruence has been passed.
+    for (dimension_type i = cgs.num_rows(); i-- > 0; )
+      if (cgs[i].is_inconsistent()) {
+        // Inconsistent congruence found: the grid is empty.
+        status.set_empty();
+        // Insert the zero dim false congruence system into `con_sys'.
+        // `gen_sys' is already in empty form.
+        con_sys.insert(Congruence::zero_dim_false());
+        PPL_ASSERT(OK());
+        return;
+      }
     set_zero_dim_univ();
   }
   PPL_ASSERT(OK());
@@ -424,12 +422,8 @@ PPL::Grid::max_min(const Linear_Expression& expr,
     exact_div_assign(ext_d, ext_d, gcd);
 
     included = true;
-    if (point != 0) {
-      Linear_Expression e;
-      for (dimension_type i = space_dim; i-- > 0; )
-	e += gen.coefficient(Variable(i)) * Variable(i);
-      *point = Generator::point(e, gen.divisor());
-    }
+    if (point != 0)
+      *point = Generator::point(Linear_Expression(gen), gen.divisor());
     return true;
   }
   return false;
@@ -455,8 +449,8 @@ PPL::Grid::set_empty() {
   // Extend the zero dim false congruence system to the appropriate
   // dimension and then swap it with `con_sys'.
   Congruence_System cgs(Congruence::zero_dim_false());
-  cgs.increase_space_dimension(space_dim);
-  con_sys.m_swap(cgs);
+  cgs.set_space_dimension(space_dim);
+  swap(con_sys, cgs);
 }
 
 void
@@ -581,13 +575,13 @@ PPL::Grid::normalize_divisors(Grid_Generator_System& sys,
     // gen_sys should have at least one point.
     PPL_ASSERT(row < num_rows);
   }
-  Grid_Generator& first_point = gen_sys[row];
+  const Grid_Generator& first_point = gen_sys[row];
   const Coefficient& gen_sys_divisor = first_point.divisor();
 
 #ifndef NDEBUG
   // Check that the divisors in gen_sys are equal.
   for (dimension_type i = row + 1; i < num_rows; ++i) {
-    Grid_Generator& g = gen_sys[i];
+    const Grid_Generator& g = gen_sys[i];
     if (g.is_parameter_or_point())
       PPL_ASSERT(gen_sys_divisor == g.divisor());
   }
@@ -612,7 +606,6 @@ PPL::Grid::normalize_divisors(Grid_Generator_System& sys,
 			      const Grid_Generator* first_point) {
   PPL_ASSERT(divisor >= 0);
   if (sys.space_dimension() > 0 && divisor > 0) {
-    dimension_type row = 0;
     dimension_type num_rows = sys.num_rows();
 
     if (first_point != 0)
@@ -620,6 +613,7 @@ PPL::Grid::normalize_divisors(Grid_Generator_System& sys,
     else {
       PPL_ASSERT(num_rows > 0);
       // Move to the first point or parameter.
+      dimension_type row = 0;
       while (sys[row].is_line())
 	if (++row == num_rows)
 	  // All rows are lines.
@@ -637,8 +631,11 @@ PPL::Grid::normalize_divisors(Grid_Generator_System& sys,
 
     // Represent every point and every parameter using the newly
     // calculated divisor.
-    for (row = num_rows; row-- > 0; )
-      sys[row].scale_to_divisor(divisor);
+    for (dimension_type i = num_rows; i-- > 0; )
+      sys.sys.rows[i].scale_to_divisor(divisor);
+
+    // Put the rows back into the linear system.
+    PPL_ASSERT(sys.sys.OK());
   }
 }
 
@@ -704,8 +701,7 @@ PPL::Grid::refine_no_check(const Constraint& c) {
 }
 
 void
-PPL::Grid::throw_invalid_argument(const char* method,
-				  const char* reason) const {
+PPL::Grid::throw_invalid_argument(const char* method, const char* reason) {
   std::ostringstream s;
   s << "PPL::Grid::" << method << ":" << std::endl
     << reason << ".";
@@ -732,9 +728,9 @@ PPL::Grid::throw_dimension_incompatible(const char* method,
 
 void
 PPL::Grid::throw_dimension_incompatible(const char* method,
-					const char* e_name,
-					const Linear_Expression& e) const {
-  throw_dimension_incompatible(method, e_name, e.space_dimension());
+					const char* le_name,
+					const Linear_Expression& le) const {
+  throw_dimension_incompatible(method, le_name, le.space_dimension());
 }
 
 void
@@ -810,7 +806,7 @@ throw_dimension_incompatible(const char* method,
 
 void
 PPL::Grid::throw_invalid_constraint(const char* method,
-				    const char* c_name) const {
+				    const char* c_name) {
   std::ostringstream s;
   s << "PPL::Grid::" << method << ":" << std::endl
     << c_name << " is not an equality constraint.";
@@ -819,7 +815,7 @@ PPL::Grid::throw_invalid_constraint(const char* method,
 
 void
 PPL::Grid::throw_invalid_constraints(const char* method,
-				    const char* cs_name) const {
+				    const char* cs_name) {
   std::ostringstream s;
   s << "PPL::Grid::" << method << ":" << std::endl
     << "the constraint system " << cs_name
@@ -829,7 +825,7 @@ PPL::Grid::throw_invalid_constraints(const char* method,
 
 void
 PPL::Grid::throw_invalid_generator(const char* method,
-				   const char* g_name) const {
+				   const char* g_name) {
   std::ostringstream s;
   s << "PPL::Grid::" << method << ":" << std::endl
     << "*this is an empty grid and "
@@ -839,7 +835,7 @@ PPL::Grid::throw_invalid_generator(const char* method,
 
 void
 PPL::Grid::throw_invalid_generators(const char* method,
-				    const char* gs_name) const {
+				    const char* gs_name) {
   std::ostringstream s;
   s << "PPL::Grid::" << method << ":" << std::endl
     << "*this is an empty grid and" << std::endl
