@@ -25,6 +25,7 @@ site: http://bugseng.com/products/ppl/ . */
 #include "Grid.defs.hh"
 #include "Topology.types.hh"
 #include "Scalar_Products.defs.hh"
+#include "Scalar_Products.inlines.hh"
 #include "Polyhedron.defs.hh"
 #include "assert.hh"
 #include <iostream>
@@ -48,7 +49,7 @@ PPL::Grid::Grid(const Grid& y, Complexity_Class)
     if (y.congruences_are_up_to_date())
       con_sys = y.con_sys;
     else
-      con_sys.increase_space_dimension(space_dim);
+      con_sys.set_space_dimension(space_dim);
     if (y.generators_are_up_to_date())
       gen_sys = y.gen_sys;
     else
@@ -85,8 +86,7 @@ PPL::Grid::Grid(const Constraint_System& cs)
     return;
   }
 
-  Congruence_System cgs;
-  cgs.insert(0*Variable(space_dim - 1) %= 1);
+  Congruence_System cgs(cs.space_dimension());
   for (Constraint_System::const_iterator i = cs.begin(),
 	 cs_end = cs.end(); i != cs_end; ++i)
     if (i->is_equality())
@@ -125,8 +125,7 @@ PPL::Grid::Grid(Constraint_System& cs, Recycle_Input)
     return;
   }
 
-  Congruence_System cgs;
-  cgs.insert(0*Variable(space_dim - 1) %= 1);
+  Congruence_System cgs(space_dim);
   for (Constraint_System::const_iterator i = cs.begin(),
 	 cs_end = cs.end(); i != cs_end; ++i)
     if (i->is_equality())
@@ -178,8 +177,7 @@ PPL::Grid::Grid(const Polyhedron& ph,
     // Only the equality constraints need be used.
     PPL_ASSERT(ph.constraints_are_up_to_date());
     const Constraint_System& cs = ph.constraints();
-    Congruence_System cgs;
-    cgs.insert(0*Variable(space_dim - 1) %= 1);
+    Congruence_System cgs(space_dim);
     for (Constraint_System::const_iterator i = cs.begin(),
            cs_end = cs.end(); i != cs_end; ++i)
       if (i->is_equality())
@@ -193,15 +191,14 @@ PPL::Grid::Grid(const Polyhedron& ph,
     const Generator_System& gs = ph.generators();
     Grid_Generator_System ggs(space_dim);
     Linear_Expression point_expr;
+    point_expr.set_space_dimension(space_dim);
     PPL_DIRTY_TEMP_COEFFICIENT(point_divisor);
     for (Generator_System::const_iterator g = gs.begin(),
            gs_end = gs.end(); g != gs_end; ++g) {
       if (g->is_point() || g->is_closure_point()) {
-        for (dimension_type i = space_dim; i-- > 0; ) {
-          const Variable v(i);
-          point_expr += g->coefficient(v) * v;
-          point_divisor = g->divisor();
-        }
+        point_expr.linear_combine(g->expr, Coefficient_one(), Coefficient_one(),
+                                  1, space_dim + 1);
+        point_divisor = g->divisor();
         ggs.insert(grid_point(point_expr, point_divisor));
         break;
       }
@@ -210,27 +207,22 @@ PPL::Grid::Grid(const Polyhedron& ph,
     // If the polyhedron's generator is a (closure) point, the grid line must
     // have the direction given by a line that joins the grid point already
     // inserted and the new point.
-    PPL_DIRTY_TEMP_COEFFICIENT(coeff);
-    PPL_DIRTY_TEMP_COEFFICIENT(g_divisor);
     for (Generator_System::const_iterator g = gs.begin(),
            gs_end = gs.end(); g != gs_end; ++g) {
       Linear_Expression e;
+      e.set_space_dimension(space_dim);
       if (g->is_point() || g->is_closure_point()) {
-        g_divisor = g->divisor();
-        for (dimension_type i = space_dim; i-- > 0; ) {
-          const Variable v(i);
-          coeff = point_expr.coefficient(v) * g_divisor;
-          coeff -= g->coefficient(v) * point_divisor;
-          e += coeff * v;
-        }
+        e.linear_combine(point_expr, Coefficient_one(), g->divisor(),
+                         1, space_dim + 1);
+        e.linear_combine(g->expr, Coefficient_one(), -point_divisor,
+                         1, space_dim + 1);
         if (e.all_homogeneous_terms_are_zero())
           continue;
       }
-      else
-        for (dimension_type i = space_dim; i-- > 0; ) {
-          const Variable v(i);
-          e += g->coefficient(v) * v;
-        }
+      else {
+        e.linear_combine(g->expr, Coefficient_one(), Coefficient_one(),
+                         1, space_dim + 1);
+      }
       ggs.insert(grid_line(e));
     }
     construct(ggs);
@@ -284,7 +276,7 @@ PPL::Grid::congruences() const {
 
   if (space_dim == 0) {
     // Zero-dimensional universe.
-    PPL_ASSERT(con_sys.num_rows() == 0 && con_sys.num_columns() == 2);
+    PPL_ASSERT(con_sys.num_rows() == 0 && con_sys.space_dimension() == 0);
     return con_sys;
   }
 
@@ -553,11 +545,7 @@ PPL::Grid::relation_with(const Generator& g) const {
   if (!congruences_are_up_to_date())
     update_congruences();
 
-  Linear_Expression expr;
-  for (dimension_type i = g_space_dim; i-- > 0; ) {
-    const Variable v(i);
-    expr += g.coefficient(v) * v;
-  }
+  Linear_Expression expr(g);
   Grid_Generator gg(grid_point());
   if (g.is_point() || g.is_closure_point())
     // Points and closure points are converted to grid points.
@@ -644,21 +632,11 @@ PPL::Grid::relation_with(const Constraint& c) const {
 	// and fall through into the parameter case.
 	Grid_Generator& gen = const_cast<Grid_Generator&>(g);
 	const Grid_Generator& point = *first_point;
-        const Coefficient& p_div = g[0];
-        if (p_div != Coefficient_one()) {
-          for (dimension_type j = gen.size() - 1; j-- > 1; )
-            gen[j] *= p_div;
-        }
-        const Coefficient& g_div = gen[0];
-        if (g_div == Coefficient_one()) {
-          for (dimension_type j = gen.size() - 1; j-- > 1; )
-            gen[j] -= point[j];
-        }
-        else {
-          for (dimension_type j = gen.size() - 1; j-- > 1; )
-            sub_mul_assign(gen[j], g_div, point[j]);
-        }
-        gen[0] *= p_div;
+        const Coefficient& p_div = point.divisor();
+        const Coefficient& g_div = gen.divisor();
+        gen.expr.linear_combine(point.expr, p_div, -g_div,
+                                1, gen.expr.space_dimension());
+        gen.expr.set_inhomogeneous_term(g_div * p_div);
         gen.strong_normalize();
 	gen.set_is_parameter();
         PPL_ASSERT(gen.OK());
@@ -669,8 +647,8 @@ PPL::Grid::relation_with(const Constraint& c) const {
     case Grid_Generator::LINE:
       {
         const int sign = c.is_strict_inequality()
-          ? Scalar_Products::reduced_sign(c, g)
-          : Scalar_Products::sign(c, g);
+          ? Scalar_Products::reduced_sign(c.expr, g.expr)
+          : Scalar_Products::sign(c.expr, g.expr);
         if (sign != 0)
           return Poly_Con_Relation::strictly_intersects();
       }
@@ -736,11 +714,18 @@ PPL::Grid::is_universe() const {
   // Test con_sys's inclusion in a universe generator system.
 
   // The zero dimension cases are handled above.
-  Variable var(space_dim - 1);
-  for (dimension_type i = space_dim; i-- > 0; )
-    if (!con_sys.satisfies_all_congruences(grid_line(Variable(i) + var)))
+  for (dimension_type i = space_dim; i-- > 0; ) {
+    Linear_Expression expr;
+    expr.set_space_dimension(space_dim);
+    expr += Variable(i);
+    if (!con_sys.satisfies_all_congruences(grid_line(expr)))
       return false;
-  PPL_ASSERT(con_sys.satisfies_all_congruences(grid_point(0*var)));
+  }
+#ifndef NDEBUG
+  Linear_Expression expr;
+  expr.set_space_dimension(space_dim);
+  PPL_ASSERT(con_sys.satisfies_all_congruences(grid_point(expr)));
+#endif
   return true;
 }
 
@@ -846,19 +831,16 @@ PPL::Grid::constrains(const Variable var) const {
     }
 
     // Scan generators: perhaps we will find line(var).
-    const dimension_type var_id = var.id();
     for (dimension_type i = gen_sys.num_rows(); i-- > 0; ) {
       const Grid_Generator& g_i = gen_sys[i];
-      if (g_i.is_line()) {
-	if (sgn(g_i.coefficient(var)) != 0) {
-	  for (dimension_type j = 0; j < space_dim; ++j)
-	    if (g_i.coefficient(Variable(j)) != 0 && j != var_id)
-	      goto next;
+      if (!g_i.is_line())
+        continue;
+      if (sgn(g_i.coefficient(var)) != 0) {
+        if (g_i.expression().all_zeroes(1, var.space_dimension())
+            && g_i.expression().all_zeroes(var.space_dimension() + 1, space_dim + 1))
+          // The only nonzero coefficient in g_i is the one of var.
           return true;
-	}
       }
-    next:
-      ;
     }
 
     // We are still here: at least we know that the grid is not empty.
@@ -944,7 +926,7 @@ PPL::Grid::OK(bool check_not_empty) const {
     // `con_sys' : number of congruences x poly_num_columns
     // `gen_sys' : number of generators  x poly_num_columns
     if (congruences_are_up_to_date())
-      if (con_sys.num_columns() != num_columns + 1 /* moduli */) {
+      if (con_sys.space_dimension() != space_dim) {
 #ifndef NDEBUG
 	cerr << "Incompatible size! (con_sys and space_dim)"
 	     << endl;
@@ -953,7 +935,7 @@ PPL::Grid::OK(bool check_not_empty) const {
       }
 
     if (generators_are_up_to_date()) {
-      if (gen_sys.space_dimension() + 1 != num_columns) {
+      if (gen_sys.space_dimension() != space_dim) {
 #ifndef NDEBUG
 	cerr << "Incompatible size! (gen_sys and space_dim)"
 	     << endl;
@@ -961,22 +943,6 @@ PPL::Grid::OK(bool check_not_empty) const {
 	goto fail;
       }
 
-      // Check if the system of generators is well-formed.
-      if (!gen_sys.OK())
-	goto fail;
-
-      // Check each generator in the system.
-      for (dimension_type i = gen_sys.num_rows(); i-- > 0; ) {
-	const Grid_Generator& g = gen_sys[i];
-
-	if (g.size() < 1) {
-#ifndef NDEBUG
-	  cerr << "Parameter should have coefficients." << endl;
-#endif
-	  goto fail;
-	}
-      }
-
       // A non-empty system of generators describing a grid is valid
       // if and only if it contains a point.
       if (!gen_sys.has_no_rows() && !gen_sys.has_points()) {
@@ -1038,7 +1004,7 @@ PPL::Grid::OK(bool check_not_empty) const {
 	// contain at least a single point afterward.
 	PPL_ASSERT(!gs.has_no_rows());
 	for (dimension_type row = gen_sys.num_rows(); row-- > 0; ) {
-	  Grid_Generator& g = gs[row];
+	  const Grid_Generator& g = gs[row];
 	  const Grid_Generator& g_copy = gen_sys[row];
 	  if (g.is_equal_to(g_copy))
 	    continue;
@@ -1108,7 +1074,7 @@ PPL::Grid::OK(bool check_not_empty) const {
 	goto fail;
       }
 
-      if (dim_kinds.size() != con_sys.num_columns() - 1 /* modulus */) {
+      if (dim_kinds.size() != con_sys.space_dimension() + 1 /* inhomogeneous term */) {
 #ifndef NDEBUG
 	cerr << "Size of dim_kinds should equal the number of columns."
 	     << endl;
@@ -1234,7 +1200,7 @@ PPL::Grid::add_recycled_congruences(Congruence_System& cgs) {
 
   // Swap (instead of copying) the coefficients of `cgs' (which is
   // writable).
-  con_sys.recycling_insert(cgs);
+  con_sys.insert(cgs, Recycle_Input());
 
   // Congruences may not be minimized and generators are out of date.
   clear_congruences_minimized();
@@ -1274,7 +1240,7 @@ PPL::Grid::add_recycled_grid_generators(Grid_Generator_System& gs) {
       update_generators();
     normalize_divisors(gs, gen_sys);
 
-    gen_sys.recycling_insert(gs);
+    gen_sys.insert(gs, Recycle_Input());
 
     // Congruences are out of date and generators are not minimized.
     clear_congruences_up_to_date();
@@ -1291,7 +1257,7 @@ PPL::Grid::add_recycled_grid_generators(Grid_Generator_System& gs) {
     throw_invalid_generators("add_recycled_grid_generators(gs)", "gs");
 
   // Adjust `gs' to the right dimension.
-  gs.insert(parameter(0*Variable(space_dim-1)));
+  gs.set_space_dimension(space_dim);
 
   gen_sys.m_swap(gs);
 
@@ -1346,7 +1312,7 @@ PPL::Grid::unconstrain(const Variable var) {
 
   PPL_ASSERT(generators_are_up_to_date());
   Grid_Generator l = grid_line(var);
-  gen_sys.recycling_insert(l);
+  gen_sys.insert(l, Recycle_Input());
   // With the added generator, congruences are out of date.
   clear_congruences_up_to_date();
   clear_generators_minimized();
@@ -1378,7 +1344,7 @@ PPL::Grid::unconstrain(const Variables_Set& vars) {
   for (Variables_Set::const_iterator vsi = vars.begin(),
          vsi_end = vars.end(); vsi != vsi_end; ++vsi) {
     Grid_Generator l = grid_line(Variable(*vsi));
-    gen_sys.recycling_insert(l);
+    gen_sys.insert(l, Recycle_Input());
   }
   // Constraints are no longer up-to-date.
   clear_generators_minimized();
@@ -1456,7 +1422,7 @@ PPL::Grid::upper_bound_assign(const Grid& y) {
   // Match the divisors of the x and y generator systems.
   Grid_Generator_System gs(y.gen_sys);
   normalize_divisors(x.gen_sys, gs);
-  x.gen_sys.recycling_insert(gs);
+  x.gen_sys.insert(gs, Recycle_Input());
   // Congruences may be out of date and generators may have lost
   // minimal form.
   x.clear_congruences_up_to_date();
@@ -1628,10 +1594,8 @@ PPL::Grid::simplify_using_context_assign(const Grid& y) {
 	  if (y_modulus_i > 1)
 	    gr.refine_no_check(le == 1);
 	  else {
-	    Linear_Expression le2;
-	    for (dimension_type j = le.space_dimension(); j-- > 0; )
-	      le2 += 2 * le.coefficient(Variable(j)) * Variable(j);
-	    le2 += 2 * le.inhomogeneous_term();
+	    Linear_Expression le2 = le;
+            le2 *= 2;
 	    gr.refine_no_check(le2 == y_modulus_i);
 	  }
 	  break;
@@ -1792,15 +1756,18 @@ PPL::Grid::affine_image(const Variable var,
   if (marked_empty())
     return;
 
-  if (var_space_dim <= expr_space_dim && expr[var_space_dim] != 0) {
+  Coefficient_traits::const_reference expr_var = expr.coefficient(var);
+
+  if (var_space_dim <= expr_space_dim
+      && expr_var != 0) {
     // The transformation is invertible.
     if (generators_are_up_to_date()) {
       // Grid_Generator_System::affine_image() requires the third argument
       // to be a positive Coefficient.
       if (denominator > 0)
-	gen_sys.affine_image(var_space_dim, expr, denominator);
+	gen_sys.affine_image(var, expr, denominator);
       else
-	gen_sys.affine_image(var_space_dim, -expr, -denominator);
+	gen_sys.affine_image(var, -expr, -denominator);
       clear_generators_minimized();
       // Strong normalization in gs::affine_image may have modified
       // divisors.
@@ -1811,19 +1778,18 @@ PPL::Grid::affine_image(const Variable var,
       // after copying and negating `expr',
       // we exchange the roles of `expr[var_space_dim]' and `denominator'.
       Linear_Expression inverse;
-      if (expr[var_space_dim] > 0) {
+      if (expr_var > 0) {
 	inverse = -expr;
-	inverse[var_space_dim] = denominator;
-	con_sys.affine_preimage(var_space_dim, inverse, expr[var_space_dim]);
+	inverse.set_coefficient(var, denominator);
+	con_sys.affine_preimage(var, inverse, expr_var);
       }
       else {
 	// The new denominator is negative: we negate everything once
 	// more, as Congruence_System::affine_preimage() requires the
 	// third argument to be positive.
 	inverse = expr;
-	inverse[var_space_dim] = denominator;
-	neg_assign(inverse[var_space_dim]);
-	con_sys.affine_preimage(var_space_dim, inverse, -expr[var_space_dim]);
+	inverse.set_coefficient(var, -denominator);
+	con_sys.affine_preimage(var, inverse, -expr_var);
       }
       clear_congruences_minimized();
     }
@@ -1837,9 +1803,9 @@ PPL::Grid::affine_image(const Variable var,
       // Grid_Generator_System::affine_image() requires the third argument
       // to be a positive Coefficient.
       if (denominator > 0)
-	gen_sys.affine_image(var_space_dim, expr, denominator);
+	gen_sys.affine_image(var, expr, denominator);
       else
-	gen_sys.affine_image(var_space_dim, -expr, -denominator);
+	gen_sys.affine_image(var, -expr, -denominator);
 
       clear_congruences_up_to_date();
       clear_generators_minimized();
@@ -1874,15 +1840,17 @@ affine_preimage(const Variable var,
   if (marked_empty())
     return;
 
-  if (var_space_dim <= expr_space_dim && expr[var_space_dim] != 0) {
+  Coefficient_traits::const_reference expr_var = expr.coefficient(var);
+
+  if (var_space_dim <= expr_space_dim && expr_var != 0) {
     // The transformation is invertible.
     if (congruences_are_up_to_date()) {
       // Congruence_System::affine_preimage() requires the third argument
       // to be a positive Coefficient.
       if (denominator > 0)
-	con_sys.affine_preimage(var_space_dim, expr, denominator);
+	con_sys.affine_preimage(var, expr, denominator);
       else
-	con_sys.affine_preimage(var_space_dim, -expr, -denominator);
+	con_sys.affine_preimage(var, -expr, -denominator);
       clear_congruences_minimized();
     }
     if (generators_are_up_to_date()) {
@@ -1890,19 +1858,18 @@ affine_preimage(const Variable var,
       // after copying and negating `expr',
       // we exchange the roles of `expr[var_space_dim]' and `denominator'.
       Linear_Expression inverse;
-      if (expr[var_space_dim] > 0) {
+      if (expr_var > 0) {
 	inverse = -expr;
-	inverse[var_space_dim] = denominator;
-	gen_sys.affine_image(var_space_dim, inverse, expr[var_space_dim]);
+	inverse.set_coefficient(var, denominator);
+	gen_sys.affine_image(var, inverse, expr_var);
       }
       else {
 	// The new denominator is negative: we negate everything once
 	// more, as Grid_Generator_System::affine_image() requires the
 	// third argument to be positive.
 	inverse = expr;
-	inverse[var_space_dim] = denominator;
-	neg_assign(inverse[var_space_dim]);
-	gen_sys.affine_image(var_space_dim, inverse, -expr[var_space_dim]);
+        inverse.set_coefficient(var, -denominator);
+	gen_sys.affine_image(var, inverse, -expr_var);
       }
       clear_generators_minimized();
     }
@@ -1915,9 +1882,9 @@ affine_preimage(const Variable var,
     // Congruence_System::affine_preimage() requires the third argument
     // to be a positive Coefficient.
     if (denominator > 0)
-      con_sys.affine_preimage(var_space_dim, expr, denominator);
+      con_sys.affine_preimage(var, expr, denominator);
     else
-      con_sys.affine_preimage(var_space_dim, -expr, -denominator);
+      con_sys.affine_preimage(var, -expr, -denominator);
 
     clear_generators_up_to_date();
     clear_congruences_minimized();
@@ -1932,7 +1899,6 @@ generalized_affine_image(const Variable var,
 			 const Linear_Expression& expr,
 			 Coefficient_traits::const_reference denominator,
 			 Coefficient_traits::const_reference modulus) {
-
   // The denominator cannot be zero.
   if (denominator == 0)
     throw_invalid_argument("generalized_affine_image(v, r, e, d, m)",
@@ -1950,6 +1916,10 @@ generalized_affine_image(const Variable var,
   if (space_dim < var_space_dim)
     throw_dimension_incompatible("generalized_affine_image(v, r, e, d, m)",
 				 "v", var);
+  // The relation symbol cannot be a disequality.
+  if (relsym == NOT_EQUAL)
+    throw_invalid_argument("generalized_affine_image(v, r, e, d, m)",
+                           "r is the disequality relation symbol");
 
   // Any image of an empty grid is empty.
   if (marked_empty())
@@ -2015,20 +1985,24 @@ generalized_affine_preimage(const Variable var,
 			    Coefficient_traits::const_reference modulus) {
   // The denominator cannot be zero.
   if (denominator == 0)
-    throw_invalid_argument("generalized_affine_preimage(v, e, d, m)",
+    throw_invalid_argument("generalized_affine_preimage(v, r, e, d, m)",
 			   "d == 0");
 
   // The dimension of `expr' should be at most the dimension of
   // `*this'.
   const dimension_type expr_space_dim = expr.space_dimension();
   if (space_dim < expr_space_dim)
-    throw_dimension_incompatible("generalized_affine_preimage(v, e, d, m)",
+    throw_dimension_incompatible("generalized_affine_preimage(v, r, e, d, m)",
 				 "e", expr);
   // `var' should be one of the dimensions of the grid.
   const dimension_type var_space_dim = var.space_dimension();
   if (space_dim < var_space_dim)
-    throw_dimension_incompatible("generalized_affine_preimage(v, e, d, m)",
+    throw_dimension_incompatible("generalized_affine_preimage(v, r, e, d, m)",
 				 "v", var);
+  // The relation symbol cannot be a disequality.
+  if (relsym == NOT_EQUAL)
+    throw_invalid_argument("generalized_affine_preimage(v, r, e, d, m)",
+                           "r is the disequality relation symbol");
 
   // If relsym is not EQUAL, then we return a safe approximation
   // by adding a line in the direction of var.
@@ -2110,14 +2084,18 @@ generalized_affine_image(const Linear_Expression& lhs,
   // `*this'.
   dimension_type lhs_space_dim = lhs.space_dimension();
   if (space_dim < lhs_space_dim)
-    throw_dimension_incompatible("generalized_affine_image(e1, r, e2)",
+    throw_dimension_incompatible("generalized_affine_image(e1, r, e2, m)",
 				 "e1", lhs);
   // The dimension of `rhs' should be at most the dimension of
   // `*this'.
   const dimension_type rhs_space_dim = rhs.space_dimension();
   if (space_dim < rhs_space_dim)
-    throw_dimension_incompatible("generalized_affine_image(e1, r, e2)",
+    throw_dimension_incompatible("generalized_affine_image(e1, r, e2, m)",
 				 "e2", rhs);
+  // The relation symbol cannot be a disequality.
+  if (relsym == NOT_EQUAL)
+    throw_invalid_argument("generalized_affine_image(e1, r, e2, m)",
+                           "r is the disequality relation symbol");
 
   // Any image of an empty grid is empty.
   if (marked_empty())
@@ -2138,9 +2116,9 @@ generalized_affine_image(const Linear_Expression& lhs,
     if (marked_empty())
       return;
 
-    for (dimension_type i = space_dim; i-- > 0; )
-      if (lhs.coefficient(Variable(i)) != 0)
-	add_grid_generator(grid_line(Variable(i)));
+    for (Linear_Expression::const_iterator i = lhs.begin(), i_end = lhs.end();
+          i != i_end; ++i)
+      add_grid_generator(grid_line(i.variable()));
 
     PPL_ASSERT(OK());
     return;
@@ -2155,29 +2133,23 @@ generalized_affine_image(const Linear_Expression& lhs,
 
   // Compute the actual space dimension of `lhs',
   // i.e., the highest dimension having a non-zero coefficient in `lhs'.
-  do {
-    if (lhs_space_dim == 0) {
-      // All variables have zero coefficients, so `lhs' is a constant.
-      add_congruence_no_check((lhs %= rhs) / tmp_modulus);
-      return;
-    }
+
+  lhs_space_dim = lhs.last_nonzero();
+  if (lhs_space_dim == 0) {
+    // All variables have zero coefficients, so `lhs' is a constant.
+    add_congruence_no_check((lhs %= rhs) / tmp_modulus);
+    return;
   }
-  while (lhs.coefficient(Variable(--lhs_space_dim)) == 0);
 
   // Gather in `new_lines' the collections of all the lines having the
-  // direction of variables occurring in `lhs'.  While at it, check
-  // whether there exists a variable occurring in both `lhs' and
-  // `rhs'.
+  // direction of variables occurring in `lhs'.
   Grid_Generator_System new_lines;
-  bool lhs_vars_intersect_rhs_vars = false;
-  for (dimension_type i = lhs_space_dim + 1; i-- > 0; )
-    if (lhs.coefficient(Variable(i)) != 0) {
-      new_lines.insert(grid_line(Variable(i)));
-      if (rhs.coefficient(Variable(i)) != 0)
-	lhs_vars_intersect_rhs_vars = true;
-    }
+  for (Linear_Expression::const_iterator i = lhs.begin(),
+        i_end = lhs.lower_bound(Variable(lhs_space_dim)); i != i_end; ++i)
+    new_lines.insert(grid_line(i.variable()));
 
-  if (lhs_vars_intersect_rhs_vars) {
+  dimension_type num_common_dims = std::min(lhs_space_dim, rhs_space_dim);
+  if (lhs.have_a_common_variable(rhs, Variable(0), Variable(num_common_dims))) {
     // Some variables in `lhs' also occur in `rhs'.
     // To ease the computation, add an additional dimension.
     const Variable new_var(space_dim);
@@ -2194,10 +2166,10 @@ generalized_affine_image(const Linear_Expression& lhs,
       // hand side expression.
 
       // Adjust `new_lines' to the right dimension.
-      new_lines.insert(parameter(0*Variable(space_dim-1)));
+      new_lines.set_space_dimension(space_dim);
       // Add the lines to `gen_sys' (first make sure they are up-to-date).
       update_generators();
-      gen_sys.recycling_insert(new_lines);
+      gen_sys.insert(new_lines, Recycle_Input());
       normalize_divisors(gen_sys);
       // Update the flags.
       clear_congruences_up_to_date();
@@ -2239,17 +2211,20 @@ generalized_affine_preimage(const Linear_Expression& lhs,
 			    const Relation_Symbol relsym,
 			    const Linear_Expression& rhs,
 			    Coefficient_traits::const_reference modulus) {
-
   // The dimension of `lhs' must be at most the dimension of `*this'.
   dimension_type lhs_space_dim = lhs.space_dimension();
   if (space_dim < lhs_space_dim)
-    throw_dimension_incompatible("generalized_affine_preimage(e1, e2, m)",
+    throw_dimension_incompatible("generalized_affine_preimage(e1, r, e2, m)",
 				 "lhs", lhs);
   // The dimension of `rhs' must be at most the dimension of `*this'.
   const dimension_type rhs_space_dim = rhs.space_dimension();
   if (space_dim < rhs_space_dim)
-    throw_dimension_incompatible("generalized_affine_preimage(e1, e2, m)",
+    throw_dimension_incompatible("generalized_affine_preimage(e1, r, e2, m)",
 				 "e2", rhs);
+  // The relation symbol cannot be a disequality.
+  if (relsym == NOT_EQUAL)
+    throw_invalid_argument("generalized_affine_preimage(e1, r, e2, m)",
+                           "r is the disequality relation symbol");
 
   // Any preimage of an empty grid is empty.
   if (marked_empty())
@@ -2270,9 +2245,9 @@ generalized_affine_preimage(const Linear_Expression& lhs,
     if (marked_empty())
       return;
 
-    for (dimension_type i = lhs_space_dim + 1; i-- > 0; )
-      if (lhs.coefficient(Variable(i)) != 0)
-	add_grid_generator(grid_line(Variable(i)));
+    for (Linear_Expression::const_iterator i = lhs.begin(), i_end = lhs.end();
+          i != i_end; ++i)
+      add_grid_generator(grid_line(i.variable()));
 
     PPL_ASSERT(OK());
     return;
@@ -2287,30 +2262,23 @@ generalized_affine_preimage(const Linear_Expression& lhs,
 
   // Compute the actual space dimension of `lhs',
   // i.e., the highest dimension having a non-zero coefficient in `lhs'.
-  do {
-    if (lhs_space_dim == 0) {
-      // All variables have zero coefficients, so `lhs' is a constant.
-      // In this case, preimage and image happen to be the same.
-      add_congruence_no_check((lhs %= rhs) / tmp_modulus);
-      return;
-    }
+  lhs_space_dim = lhs.last_nonzero();
+  if (lhs_space_dim == 0) {
+    // All variables have zero coefficients, so `lhs' is a constant.
+    // In this case, preimage and image happen to be the same.
+    add_congruence_no_check((lhs %= rhs) / tmp_modulus);
+    return;
   }
-  while (lhs.coefficient(Variable(--lhs_space_dim)) == 0);
 
   // Gather in `new_lines' the collections of all the lines having
   // the direction of variables occurring in `lhs'.
-  // While at it, check whether or not there exists a variable
-  // occurring in both `lhs' and `rhs'.
   Grid_Generator_System new_lines;
-  bool lhs_vars_intersect_rhs_vars = false;
-  for (dimension_type i = lhs_space_dim + 1; i-- > 0; )
-    if (lhs.coefficient(Variable(i)) != 0) {
-      new_lines.insert(grid_line(Variable(i)));
-      if (rhs.coefficient(Variable(i)) != 0)
-	lhs_vars_intersect_rhs_vars = true;
-    }
+  for (Linear_Expression::const_iterator i = lhs.begin(),
+        i_end = lhs.lower_bound(Variable(lhs_space_dim)); i != i_end; ++i)
+      new_lines.insert(grid_line(i.variable()));
 
-  if (lhs_vars_intersect_rhs_vars) {
+  dimension_type num_common_dims = std::min(lhs_space_dim, rhs_space_dim);
+  if (lhs.have_a_common_variable(rhs, Variable(0), Variable(num_common_dims))) {
     // Some variables in `lhs' also occur in `rhs'.
     // To ease the computation, add an additional dimension.
     const Variable new_var(space_dim);
@@ -2327,10 +2295,10 @@ generalized_affine_preimage(const Linear_Expression& lhs,
       // hand side
 
       // Adjust `new_lines' to the right dimension.
-      new_lines.insert(parameter(0*Variable(space_dim-1)));
+      new_lines.set_space_dimension(space_dim);
       // Add the lines to `gen_sys' (first make sure they are up-to-date).
       update_generators();
-      gen_sys.recycling_insert(new_lines);
+      gen_sys.insert(new_lines, Recycle_Input());
       normalize_divisors(gen_sys);
       // Update the flags.
       clear_congruences_up_to_date();
@@ -2385,11 +2353,11 @@ bounded_affine_image(const Variable var,
   // greater than the dimension of `*this'.
   const dimension_type lb_space_dim = lb_expr.space_dimension();
   if (space_dim < lb_space_dim)
-    throw_dimension_incompatible("bounded_affine_image(v, lb, ub)",
+    throw_dimension_incompatible("bounded_affine_image(v, lb, ub, d)",
 				 "lb", lb_expr);
   const dimension_type ub_space_dim = ub_expr.space_dimension();
   if (space_dim < ub_space_dim)
-    throw_dimension_incompatible("bounded_affine_image(v, lb, ub)",
+    throw_dimension_incompatible("bounded_affine_image(v, lb, ub, d)",
 				 "ub", ub_expr);
 
   // Any image of an empty grid is empty.
@@ -2428,11 +2396,11 @@ bounded_affine_preimage(const Variable var,
   // greater than the dimension of `*this'.
   const dimension_type lb_space_dim = lb_expr.space_dimension();
   if (space_dim < lb_space_dim)
-    throw_dimension_incompatible("bounded_affine_preimage(v, lb, ub)",
+    throw_dimension_incompatible("bounded_affine_preimage(v, lb, ub, d)",
 				 "lb", lb_expr);
   const dimension_type ub_space_dim = ub_expr.space_dimension();
   if (space_dim < ub_space_dim)
-    throw_dimension_incompatible("bounded_affine_preimage(v, lb, ub)",
+    throw_dimension_incompatible("bounded_affine_preimage(v, lb, ub, d)",
 				 "ub", ub_expr);
 
   // Any preimage of an empty grid is empty.
@@ -2480,12 +2448,14 @@ PPL::Grid::time_elapse_assign(const Grid& y) {
   normalize_divisors(gs, gen_sys);
 
   for (dimension_type i = gs_num_rows; i-- > 0; ) {
-    Grid_Generator& g = gs[i];
+    Grid_Generator& g = gs.sys.rows[i];
     if (g.is_point())
       // Transform the point into a parameter.
       g.set_is_parameter();
   }
 
+  PPL_ASSERT(gs.sys.OK());
+
   if (gs_num_rows == 0)
     // `y' was the grid containing a single point at the origin, so
     // the result is `x'.
@@ -2493,7 +2463,7 @@ PPL::Grid::time_elapse_assign(const Grid& y) {
 
   // Append `gs' to the generators of `x'.
 
-  gen_sys.recycling_insert(gs);
+  gen_sys.insert(gs, Recycle_Input());
 
   x.clear_congruences_up_to_date();
   x.clear_generators_minimized();
